import 'package:flutter/material.dart';
import 'package:date_util/date_util.dart';

final _dateUtil = DateUtil();

int _dayInMonth(int row, int column, DateTime firstDay) => ((row * 7) + (column + 1)) - (firstDay.weekday + 1) + 1;
int _neededRows(DateTime firstDayInMonth) {
  int dim = _dateUtil.daysInMonth(firstDayInMonth.month, firstDayInMonth.year);
  dim += firstDayInMonth.weekday - 1;
  return (dim / 7).ceil();
}

class DayOfWeekOptions {
  final String name;
  final Color color;
  final Color textColor;
  final Color nameTextColor;

  DayOfWeekOptions({
    this.name,
    this.color,
    this.textColor,
    this.nameTextColor,
  });
}

class DateOptions {
  final Color color;
  final Color textColor;

  DateOptions({
    this.color,
    this.textColor,
  });
}

class PanicCalendarTile<T> extends StatelessWidget {
  final int row;
  final int column;

  final DateTime selectedDay;
  final DateTime firstDay;

  final VoidCallback onTap;
  final Widget Function(BuildContext context, DateTime day, List<T> events) tileEventBuilder;

  final List<T> events;

  final Color color;
  final Color textColor;

  const PanicCalendarTile({
    Key key,
    @required this.row,
    @required this.column,
    @required this.selectedDay,
    @required this.firstDay,
    this.onTap,
    this.tileEventBuilder,
    this.events,
    this.color,
    this.textColor,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    int i = _dayInMonth(row, column, firstDay);

    Widget tileEvents = (tileEventBuilder ?? (c, d, e) => null)(
      context,
      DateTime(firstDay.year, firstDay.month, i),
      events,
    );

    if (row == 0 && column < firstDay.weekday)
      return Container();
    else if (i > _dateUtil.daysInMonth(this.firstDay.month, this.firstDay.year))
      return Container();
    else
      return Padding(
        padding: const EdgeInsets.all(2.0),
        child: FlatButton(
          onPressed: onTap,
          color: _getColor(context, i),
          child: Column(
            mainAxisAlignment: MainAxisAlignment.center,
            crossAxisAlignment: CrossAxisAlignment.center,
            mainAxisSize: MainAxisSize.max,
            children: [
              if (tileEvents != null)
                Flexible(
                  flex: 1,
                  child: IgnorePointer(
                    ignoring: true,
                    child: Opacity(
                      opacity: 0.0,
                      child: tileEvents,
                    ),
                  ),
                ),
              Flexible(
                fit: FlexFit.tight,
                child: Center(
                  child: Text(
                    i.toString(),
                    style: textColor != null ? TextStyle(color: textColor) : null,
                  ),
                ),
              ),
              if (tileEvents != null)
                Flexible(
                  flex: 1,
                  child: tileEvents,
                ),
            ],
          ),
          shape: CircleBorder(),
        ),
      );
  }

  Color _mixColors(Color c1, Color c2, double p) {
    double ip = 1 - p;

    return Color.fromARGB(
      (c1.alpha * p + c2.alpha * ip).round(),
      (c1.red * p + c2.red * ip).round(),
      (c1.green * p + c2.green * ip).round(),
      (c1.blue * p + c2.blue * ip).round(),
    );
  }

  Color _getColor(BuildContext context, int i) {
    if (color != null) {
      if (i == selectedDay.day && firstDay.month == selectedDay.month && firstDay.year == selectedDay.year)
        return _mixColors(color, Theme.of(context).primaryColor, 0.5);
      else
        return color;
    }

    if (i == selectedDay.day && firstDay.month == selectedDay.month && firstDay.year == selectedDay.year) {
      return Theme.of(context).primaryColor;
    }

    DateTime now = DateTime.now();

    if (i == now.day && firstDay.month == now.month && firstDay.year == now.year) {
      return Theme.of(context).primaryColor.withOpacity(0.5);
    }

    return null;
  }
}

class PanicCalendar<T> extends StatefulWidget {
  final Map<DateTime, List<T>> events;

  final Map<int, DayOfWeekOptions> daysOfWeek;
  final Map<int, String> monthNames;
  final Map<DateTime, DateOptions> dateOptions;

  final void Function(DateTime day) daySelected;
  final void Function(int year, int month) monthSwitched;
  final Widget Function(BuildContext context, DateTime day, List<T> events) tileEventBuilder;

  final Widget topDivider;
  final Widget bottomDivider;

  const PanicCalendar({
    Key key,
    @required this.daysOfWeek,
    this.monthNames,
    this.dateOptions,
    this.events = const {},
    this.daySelected,
    this.monthSwitched,
    this.tileEventBuilder,
    this.topDivider = const Divider(),
    this.bottomDivider = const Divider(),
  }) : super(key: key);

  @override
  _PanicCalendarState<T> createState() => _PanicCalendarState<T>();
}

class _PanicCalendarState<T> extends State<PanicCalendar<T>> {
  DateTime selectedDay;
  DateTime firstDayOfcurrentMonth;

  String get _altMonthName => _dateUtil.month(firstDayOfcurrentMonth.month);

  @override
  void initState() {
    selectedDay = DateTime.now();
    firstDayOfcurrentMonth = DateTime(selectedDay.year, selectedDay.month, 1);
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Column(
      children: [
        _buildHeader(),
        if (widget.topDivider != null) widget.topDivider,
        _buildDayNames(),
        if (widget.bottomDivider != null) widget.bottomDivider,
        for (int i = 0; i < _neededRows(firstDayOfcurrentMonth); i++)
          Expanded(
            child: _buildRow(i),
          ),
      ],
    );
  }

  Row _buildHeader() {
    return Row(
      crossAxisAlignment: CrossAxisAlignment.center,
      children: [
        IconButton(
          icon: Icon(Icons.chevron_left),
          onPressed: () {
            (widget.monthSwitched ?? (a, b) {})(firstDayOfcurrentMonth.year, firstDayOfcurrentMonth.month - 1);

            setState(() {
              firstDayOfcurrentMonth = DateTime(firstDayOfcurrentMonth.year, firstDayOfcurrentMonth.month - 1, 1);
            });
          },
        ),
        Expanded(
          child: Text(
            (widget.monthNames != null ? widget.monthNames[firstDayOfcurrentMonth.month] ?? _altMonthName : _altMonthName) + ' ${firstDayOfcurrentMonth.year}',
            textAlign: TextAlign.center,
            style: TextStyle(fontSize: 20),
          ),
        ),
        IconButton(
          icon: Icon(Icons.chevron_right),
          onPressed: () {
            (widget.monthSwitched ?? (a, b) {})(firstDayOfcurrentMonth.year, firstDayOfcurrentMonth.month + 1);

            setState(() {
              firstDayOfcurrentMonth = DateTime(firstDayOfcurrentMonth.year, firstDayOfcurrentMonth.month + 1, 1);
            });
          },
        ),
      ],
    );
  }

  Row _buildDayNames() {
    return Row(
      mainAxisAlignment: MainAxisAlignment.spaceEvenly,
      children: [
        for (int i in widget.daysOfWeek.keys)
          Expanded(
            child: Text(
              widget.daysOfWeek[i].name,
              textAlign: TextAlign.center,
              style: TextStyle(
                fontSize: 12,
                fontWeight: FontWeight.bold,
                color: widget.daysOfWeek[i].nameTextColor,
              ),
            ),
          ),
      ],
    );
  }

  Row _buildRow(int row) {
    return Row(
      mainAxisAlignment: MainAxisAlignment.spaceEvenly,
      crossAxisAlignment: CrossAxisAlignment.stretch,
      children: [
        for (int column in widget.daysOfWeek.keys)
          Expanded(
            child: _buildCalendarTile(
              row,
              column,
              DateTime(
                firstDayOfcurrentMonth.year,
                firstDayOfcurrentMonth.month,
                _dayInMonth(row, column, firstDayOfcurrentMonth),
              ),
            ),
          ),
      ],
    );
  }

  PanicCalendarTile _buildCalendarTile(int row, int column, DateTime dt) {
    Color color;
    Color textColor;

    if (widget.dateOptions != null && widget.dateOptions[dt] != null) {
      color = widget.dateOptions[dt].color;
      textColor = widget.dateOptions[dt].textColor;
    }

    if (widget.daysOfWeek != null && widget.daysOfWeek[column] != null) {
      if (color == null) color = widget.daysOfWeek[column]?.color;
      if (textColor == null) textColor = widget.daysOfWeek[column]?.textColor;
    }

    return PanicCalendarTile<T>(
      row: row,
      column: column,
      selectedDay: selectedDay,
      firstDay: firstDayOfcurrentMonth,
      tileEventBuilder: widget.tileEventBuilder,
      color: color,
      textColor: textColor,
      onTap: () {
        try {
          (widget.daySelected ?? (d) {})(dt);
        } catch (_) {}

        setState(() {
          selectedDay = dt;
        });
      },
      events: widget.events[dt],
    );
  }
}
